@@ -3,9 +3,9 @@
 from __future__ import unicode_literals
 
 try:
-    from ConfigParser import SafeConfigParser, NoOptionError
+    from configparser import RawConfigParser, NoOptionError
 except ImportError:
-    from configparser import SafeConfigParser, NoOptionError
+    from ConfigParser import RawConfigParser, NoOptionError
 
 try:
     from StringIO import StringIO
@@ -21,6 +21,18 @@ import sre_constants
 import subprocess
 import io
 from string import Formatter
+from datetime import datetime
+
+import sys
+import codecs
+sys.stdout = codecs.getwriter('utf8')(sys.stdout)
+
+__VERSION__ = '0.3.5'
+
+DESCRIPTION = 'Bumpversion: v{} (using Python v{})'.format(
+    __VERSION__,
+    sys.version.split("\n")[0].split(" ")[0],
+)
 
 
 class Git(object):
@@ -73,7 +85,7 @@ class Git(object):
         info["distance_to_latest_tag"] = int(describe_out.pop())
         info["current_version"] = describe_out.pop().lstrip("v")
 
-        #assert type(info["current_version"]) == str
+        # assert type(info["current_version"]) == str
         assert 0 == len(describe_out)
 
         return info
@@ -133,45 +145,84 @@ def prefixed_environ():
     return dict((("${}".format(key), value) for key, value in os.environ.items()))
 
 
-def attempt_version_bump(args, context):
-    #assert type(args.parse) == str
+class VersionPart(object):
 
-    try:
-        regex = re.compile(args.parse)
-    except sre_constants.error:
-        warnings.warn("--patch '{}' is not a valid regex".format(args.parse))
-        return
+    FIRST_NUMERIC = re.compile('([^\d]*)(\d+)(.*)')
 
-    if args.current_version:
-        #assert type(args.current_version) == str
-        match = regex.search(args.current_version)
-    else:
-        return
+    def __init__(self, value):
+        self.value = value
 
-    if not match:
-        warnings.warn("'{}' does not parse current version".format(args.parse))
-        return
+    def bump(self):
+        part_prefix, numeric_version, part_suffix = self.FIRST_NUMERIC.search(self.value).groups()
+        bumped_numeric = str(int(numeric_version) + 1)
+        self.value = "".join([part_prefix, bumped_numeric, part_suffix])
 
-    parsed = match.groupdict()
+    def __format__(self, format_spec):
+        return self.value
 
-    order = (label for _, label, _, _ in Formatter().parse(args.serialize))
+    def zero(self):
+        self.value = "0"
 
-    bumped = False
-    for label in order:
-        if label == args.part:
-            parsed[args.part] = int(parsed[args.part]) + 1
-            bumped = True
-        elif bumped:
-            parsed[label] = 0
 
-    parsed.update(context)
+class Version(object):
 
-    try:
-        return args.serialize.format(**parsed)
-    except KeyError as e:
-        warnings.warn("Did not find key {} in {} when serializing version number".format(
-            repr(e.message), repr(parsed)))
-        return
+    """
+    Holds a complete representation of a version string
+    """
+
+    def __init__(self, parse_regex, serialize_format, context=None):
+
+        try:
+            self.parse_regex = re.compile(parse_regex)
+        except:
+            warnings.warn("--patch '{}' is not a valid regex".format(parse_regex))
+
+        self.serialize_format = serialize_format
+
+        if not context:
+            context = {}
+
+        self.context = context
+
+    def order(self):
+        return (label for _, label, _, _ in Formatter().parse(self.serialize_format))
+
+    def register_part(self, part):
+        pass
+
+    def parse(self, version_string):
+        match = self.parse_regex.search(version_string)
+
+        self._parsed = {}
+        if not match:
+            warnings.warn("'{}' does not parse current version".format(self.parse_regex.pattern))
+            return
+
+        for key, value in match.groupdict().items():
+            self._parsed[key] = VersionPart(value)
+
+    def serialize(self):
+        values = self.context.copy()
+        values.update(self._parsed)
+        try:
+            return self.serialize_format.format(**values)
+        except KeyError as e:
+            assert hasattr(e, 'message'), dir(e)
+            warnings.warn("Did not find key {} in {} when serializing version number".format(
+                repr(e.message), repr(self._parsed)))
+            return
+
+    def bump(self, part_name):
+        bumped = False
+
+        for label in self.order():
+            if not label in self._parsed:
+                continue
+            elif label == part_name:
+                self._parsed[part_name].bump()
+                bumped = True
+            elif bumped:
+                self._parsed[label].zero()
 
 
 def main(args=None):
@@ -196,8 +247,9 @@ def main(args=None):
 
     config = None
     if os.path.exists(known_args.config_file):
-        config = SafeConfigParser()
-        config.readfp(io.open(known_args.config_file, 'rt'))
+        config = RawConfigParser()
+        config.readfp(io.open(known_args.config_file, 'rt', encoding='utf-8'))
+
         defaults.update(dict(config.items("bumpversion")))
 
         for boolvaluename in ("commit", "tag", "dry_run"):
@@ -224,7 +276,7 @@ def main(args=None):
                          default='{major}.{minor}.{patch}')
 
     parser2_2 = argparse.ArgumentParser(
-        description='Bumps version strings',
+        description=DESCRIPTION,
         formatter_class=argparse.ArgumentDefaultsHelpFormatter,
         conflict_handler='resolve',
         add_help=False,
@@ -239,16 +291,24 @@ def main(args=None):
 
     defaults.update(vars(known_args))
 
-    attempted_new_version = attempt_version_bump(
-        known_args,
-        context=dict(list(prefixed_environ().items()) + list(vcs_info.items()))
+    time_context = {
+        'now': datetime.now(),
+        'utcnow': datetime.utcnow(),
+    }
+
+    v = Version(
+        known_args.parse,
+        known_args.serialize,
+        context=dict(list(time_context.items()) + list(prefixed_environ().items()) + list(vcs_info.items()))
     )
 
-    if not ('new_version' in defaults) and attempted_new_version != None:
-        defaults['new_version'] = attempted_new_version
+    if not 'new_version' in defaults and known_args.current_version:
+        v.parse(known_args.current_version)
+        v.bump(known_args.part)
+        defaults['new_version'] = v.serialize()
 
     parser3 = argparse.ArgumentParser(
-        description='Bumps version strings',
+        description=DESCRIPTION,
         formatter_class=argparse.ArgumentDefaultsHelpFormatter,
         conflict_handler='resolve',
         parents=[parser2],
@@ -281,11 +341,11 @@ def main(args=None):
 
     parser3.add_argument('--tag-name', metavar='TAG_NAME',
                          help='Tag name (only works with --tag)',
-                         default='v{new_version}')
+                         default=defaults.get('tag_name', 'v{new_version}'))
 
     parser3.add_argument('--message', '-m', metavar='COMMIT_MSG',
                          help='Commit message',
-                         default='Bump version: {current_version} → {new_version}')
+                         default=defaults.get('message', 'Bump version: {current_version} → {new_version}'))
 
     files = []
     if 'files' in defaults:
@@ -310,19 +370,19 @@ def main(args=None):
 
     # make sure files exist and contain version string
     for path in args.files:
-        with io.open(path, 'rt') as f:
-            before = f.read()
+        with io.open(path, 'rb') as f:
+            before = f.read().decode('utf-8')
 
         assert args.current_version in before, 'Did not find string {} in file {}'.format(
             args.current_version, path)
 
     # change version string in files
     for path in args.files:
-        with io.open(path, 'rt') as f:
-            before = f.read()
+        with io.open(path, 'rb') as f:
+            before = f.read().decode('utf-8')
 
-        #assert type(args.current_version) == bytes
-        #assert type(args.new_version) == bytes
+        # assert type(args.current_version) == bytes
+        # assert type(args.new_version) == bytes
 
         after = before.replace(args.current_version, args.new_version)
 
@@ -335,19 +395,20 @@ def main(args=None):
     if config:
         config.remove_option('bumpversion', 'new_version')
 
-        if args.commit and args.tag and 'current_version' in vcs_info:
-            # we don't need to write it to the config file, because it's
-            # committed and tagged
-            config.remove_option('bumpversion', 'current_version')
-        else:
-            config.set('bumpversion', 'current_version', args.new_version)
+        config.set('bumpversion', 'current_version', args.new_version)
 
         if not args.dry_run:
             s = StringIO()
-            config.write(s)
 
-            with io.open(known_args.config_file, 'wb') as f:
-                f.write(s.getvalue().encode())
+            try:
+                config.write(s)
+                with io.open(known_args.config_file, 'wb') as f:
+                    f.write(s.getvalue().encode('utf-8'))
+            except UnicodeEncodeError:
+                warnings.warn(
+                    "Unable to write UTF-8 to config file, because of an old configparser version. "
+                    "Update with `pip install --upgrade configparser`."
+                )
 
             commit_files.append(known_args.config_file)
 
@@ -356,14 +417,14 @@ def main(args=None):
             for path in commit_files:
                 vcs.add_path(path)
 
-            formatargs = {
+            vcs_context = {
                 "current_version": args.current_version,
                 "new_version": args.new_version,
             }
-            formatargs.update(prefixed_environ())
+            vcs_context.update(time_context)
+            vcs_context.update(prefixed_environ())
 
-            vcs.commit(message=args.message.format(**formatargs))
+            vcs.commit(message=args.message.format(**vcs_context))
 
             if args.tag:
-                vcs.tag(args.tag_name.format(**formatargs))
-
+                vcs.tag(args.tag_name.format(**vcs_context))
